header {
package antlr;
}

{
import antlr.AntlrGrammarGraph;
import java.util.Enumeration;
import java.io.DataInputStream;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.*;
}

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.cs.usfca.edu
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: //depot/code/org.antlr/release/antlr-2.7.6/antlr/antlr.g#1 $
 */

class ANTLRParser extends Parser;
options {
	exportVocab=ANTLR;
	defaultErrorHandler=false;
	k=2;
  buildAST = true;
}

tokens {
	"tokens";
  RULE;
  ALTERNATIVE;
}

grammar
   :
	( 	
	 	"header"!
	 	(STRING_LITERAL!)?
	 	ACTION!
	)*
	( fileOptionsSpec! )?
	( classDef )*
	EOF!
	;

classDef
	:
	( ACTION )?!
	( DOC_COMMENT  )?!
	(!	("lexclass" | "class" id "extends" "Lexer" ) => lexerSpec
	|!	( "class" id "extends" "TreeParser" ) => treeParserSpec
	|!	parserSpec
	)
	rules
	;

fileOptionsSpec!
	:	OPTIONS
		(
			id
			ASSIGN
			optionValue
			SEMI
		)*
		RCURLY
	;

parserOptionsSpec!
  :	OPTIONS
		(
			id
			ASSIGN
			optionValue
			SEMI
		)*
		RCURLY
	;

treeParserOptionsSpec!
	:	OPTIONS
		(
			id
			ASSIGN
			optionValue
			SEMI
		)*
		RCURLY
	;

lexerOptionsSpec!
	:
	OPTIONS
	(	// Special case for vocabulary option because it has a bit-set
		"charVocabulary"
		ASSIGN
		charSet
		SEMI

	|	id
		ASSIGN
		optionValue
		SEMI
	)*
	RCURLY
	;

subruleOptionsSpec!
	:	OPTIONS
		(	id
			ASSIGN
			optionValue
			SEMI
		)*
		RCURLY
	;

// optionValue returns a Token which may be one of several things:
//    STRING_LITERAL -- a quoted string
//    CHAR_LITERAL -- a single quoted character
//		INT -- an integer
//		RULE_REF or TOKEN_REF -- an identifier
optionValue!
	:	qualifiedID
	|	STRING_LITERAL 
	|	CHAR_LTERAL 
	|	INT 
	;

charSet!
	:
		// TODO: generate a bit set
		setBlockElement
		(
			OR
			setBlockElement
		)*
	;

setBlockElement!
	:
	CHAR_LITERAL
	(
		CHAR_LITERAL
	)?
	;

tokensSpec!
	:	TOKENS
			(	(	
					TOKEN_REF
					( STRING_LITERAL )?
					(tokensSpecOptions)?
				|	STRING_LITERAL
					(tokensSpecOptions)?
				)
				SEMI
			)+
		RCURLY
	;

tokensSpecOptions!
	:	OPEN_ELEMENT_OPTION^
		id ASSIGN optionValue
		(
			SEMI
			id ASSIGN optionValue
		)*
		CLOSE_ELEMENT_OPTION
	;

superClass!
	:	LPAREN
			(STRING_LITERAL)
		RPAREN
	;

parserSpec!
	:	"class"
		id                    { 
                          AntlrGrammarGraph.currentMap = AntlrGrammarGraph.parserRules;
                          }
		(	"extends" "Parser"
			(superClass)?
		|	
		)
		SEMI
		(parserOptionsSpec)?
		(tokensSpec)?
		( ACTION )?
	;

lexerSpec!
	:   (	"lexclass"
			id                  { 
                          AntlrGrammarGraph.currentMap = AntlrGrammarGraph.lexerMap;
                          }
		|	"class"
			id                  { 
                          AntlrGrammarGraph.currentMap = AntlrGrammarGraph.lexerMap;
                          }
			"extends"
			"Lexer"
			(superClass)?
		)
		SEMI
		(lexerOptionsSpec)?
		(tokensSpec)?
		( ACTION )?
	;

treeParserSpec!
	:	"class"
		id
		"extends"
		"TreeParser"
		(superClass)?
		SEMI
		(treeParserOptionsSpec)?
		(tokensSpec)?
		( ACTION )?
	;

rules 
  : ( 
    options {
				// limitation of appox LL(k) says ambig upon
				// DOC_COMMENT TOKEN_REF, but that's an impossible sequence
				warnWhenFollowAmbig=false;
			}
		: rule  )+
  ;

rule
	:
	(	DOC_COMMENT!	)?
	(!	"protected"
	|!	"public"
	|!	"private"
	)?
	theID:id !
	( BANG! )?
	( ARG_ACTION! )?
	( "returns"! ARG_ACTION! )?
	( throwsSpec )?
	( ruleOptionsSpec )?
	(ACTION! )?
	COLON! aBlock:block SEMI!          
	( exceptionGroup )?!
  
    { 
    AntlrGrammarGraph.ruleName.add(theID_AST.getText());
    if (theID_AST.getText() != null && #rule != null && AntlrGrammarGraph.currentMap != null) {
      AntlrGrammarGraph.currentMap.put(theID_AST.getText(),#aBlock); 
    }
    #rule = #( #[RULE,"token-rule"], #rule); 
  }
	;

ruleOptionsSpec!
	:	OPTIONS
		(
			id
			ASSIGN
			optionValue
			SEMI
		)*
		RCURLY
	;

throwsSpec!
	:	"throws" id 
		( COMMA id  )*
	;

block
{
  boolean thereIsAnOR = false;
}
    :  
		alternative ( OR! {thereIsAnOR = true;} alternative )*
    
    { 
      if(thereIsAnOR) {
        #block = #( #[OR, "|"], #block);
      }
    }
    ;

alternative
    :
		( BANG! )?
		( element )* ( exceptionSpecNoLabel )?
    
    { #alternative = #( #[ALTERNATIVE, "alternative"], #alternative); }
    ;

exceptionGroup!
	:	
		( exceptionSpec )+
   ;

exceptionSpec!
   :
   "exception"
   ( ARG_ACTION )?
   ( exceptionHandler )*
   ;

exceptionSpecNoLabel!
   :
   "exception"
   ( exceptionHandler )*
   ;

exceptionHandler!
   :
   "catch"
   ARG_ACTION
   ACTION
   ;

element
	:	elementNoOptionSpec (elementOptionSpec)?
	;

elementOptionSpec!
	:	OPEN_ELEMENT_OPTION
		id ASSIGN optionValue
		(
			SEMI
			id ASSIGN optionValue
		)*
		CLOSE_ELEMENT_OPTION
	;

elementNoOptionSpec
	:!	id
		ASSIGN
		( id COLON  )?
		(	RULE_REF
			( ARG_ACTION  )?
			( BANG )?
		|	// this syntax only valid for lexer
			TOKEN_REF
			( ARG_ACTION )?
		)
	|
		(id! COLON! )?
		(	RULE_REF
			( ARG_ACTION! )?
			( BANG! )?
		|
			range 
		|
			terminal 
		|
			NOT_OP
			(	notTerminal
			|	ebnf
			)
		|
			ebnf
		)
	|
		ACTION!	
	|
		SEMPRED	
	|
		tree
	;

tree :!
	TREE_BEGIN
	rootNode
	( element )+
	RPAREN
	;

rootNode!
	:
		(id COLON )?
		terminal
//	|	range
	;

ebnf
	:	LPAREN^

		(
      options {
				warnWhenFollowAmbig = false;
        }
      : subruleOptionsSpec!
			COLON!
		|	ACTION!
			COLON!
		)?

		aBlock:block
		RPAREN!

		(	(	QUESTION ^
			|	STAR ^    
			|	PLUS ^    
			)?
			( BANG! )? 
		|
			IMPLIES	^ { #ebnf = #( #[IMPLIES] ); }
		)
	;

ast_type_spec!
	:	(	CARET 
		|	BANG 
		)?
	;

range
	:	CHAR_LITERAL RANGE CHAR_LITERAL
		( BANG )?!
	|
		(TOKEN_REF|STRING_LITERAL)
		RANGE
		(TOKEN_REF|STRING_LITERAL)
		ast_type_spec
	;

terminal
	:
		CHAR_LITERAL
		( BANG )?!
	|
		TOKEN_REF
		ast_type_spec
		// Args are only valid for lexer
		( ARG_ACTION  )?
	|
		STRING_LITERAL
		ast_type_spec
	|
		WILDCARD
		ast_type_spec
	;

notTerminal
	:
		CHAR_LITERAL
		( BANG )?!
	|
		TOKEN_REF
		ast_type_spec
	;

/** Match a.b.c.d qualified ids; WILDCARD here is overloaded as
 *  id separator; that is, I need a reference to the '.' token.
 */
qualifiedID 
	:	id
		(	WILDCARD id )*
	;

id
	:	TOKEN_REF			
	|	RULE_REF			
	;

class ANTLRLexer extends Lexer;
options {
	k=2;
	exportVocab=ANTLR;
	testLiterals=false;
	interactive=true;
	//charVocabulary='\003'..'\377';
}

tokens {
	"options";
}

{
	/**Convert 'c' to an integer char value. */
	public static int escapeCharValue(String cs) {
		//System.out.println("escapeCharValue("+cs+")");
		if ( cs.charAt(1)!='\\' ) return 0;
		switch ( cs.charAt(2) ) {
		case 'b' : return '\b';
		case 'r' : return '\r';
		case 't' : return '\t';
		case 'n' : return '\n';
		case 'f' : return '\f';
		case '"' : return '\"';
		case '\'' :return '\'';
		case '\\' :return '\\';

		case 'u' :
			// Unicode char
			if (cs.length() != 8) {
				return 0;
			}
			else {
				return
					Character.digit(cs.charAt(3), 16) * 16 * 16 * 16 +
					Character.digit(cs.charAt(4), 16) * 16 * 16 +
					Character.digit(cs.charAt(5), 16) * 16 +
					Character.digit(cs.charAt(6), 16);
			}

		case '0' :
		case '1' :
		case '2' :
		case '3' :
			if ( cs.length()>5 && Character.isDigit(cs.charAt(4)) ) {
				return (cs.charAt(2)-'0')*8*8 + (cs.charAt(3)-'0')*8 + (cs.charAt(4)-'0');
			}
			if ( cs.length()>4 && Character.isDigit(cs.charAt(3)) ) {
				return (cs.charAt(2)-'0')*8 + (cs.charAt(3)-'0');
			}
			return cs.charAt(2)-'0';

		case '4' :
		case '5' :
		case '6' :
		case '7' :
			if ( cs.length()>4 && Character.isDigit(cs.charAt(3)) ) {
				return (cs.charAt(2)-'0')*8 + (cs.charAt(3)-'0');
			}
			return cs.charAt(2)-'0';

		default :
			return 0;
		}
	}

	public static int tokenTypeForCharLiteral(String lit) {
		if ( lit.length()>3 ) {  // does char contain escape?
			return escapeCharValue(lit);
		}
		else {
			return lit.charAt(1);
		}
	}
}

WS	:	(	/*	'\r' '\n' can be matched in one alternative or by matching
				'\r' in one iteration and '\n' in another.  I am trying to
				handle any flavor of newline that comes in, but the language
				that allows both "\r\n" and "\r" and "\n" to all be valid
				newline is ambiguous.  Consequently, the resulting grammar
				must be ambiguous.  I'm shutting this warning off.
			 */
			options {
				generateAmbigWarnings=false;
			}
		:	' '
		|	'\t'
		|	'\r' '\n'	{newline();}
		|	'\r'		{newline();}
		|	'\n'		{newline();}
		)
		{ $setType(Token.SKIP); }
	;

COMMENT :
	( SL_COMMENT | t:ML_COMMENT {$setType(t.getType());} )
	{if ( _ttype != DOC_COMMENT ) $setType(Token.SKIP);}
	;

protected
SL_COMMENT :
	"//"
	( ~('\n'|'\r') )*
	(
		/*	'\r' '\n' can be matched in one alternative or by matching
			'\r' and then in the next token.  The language
			that allows both "\r\n" and "\r" and "\n" to all be valid
			newline is ambiguous.  Consequently, the resulting grammar
			must be ambiguous.  I'm shutting this warning off.
		 */
			options {
				generateAmbigWarnings=false;
			}
		:	'\r' '\n'
		|	'\r'
		|	'\n'
	)
	{ newline(); }
	;

protected
ML_COMMENT! :
	"/*"
	(	{ LA(2)!='/' }? '*' {$setType(DOC_COMMENT);}
	|
	)
	(
		/*	'\r' '\n' can be matched in one alternative or by matching
			'\r' and then in the next token.  The language
			that allows both "\r\n" and "\r" and "\n" to all be valid
			newline is ambiguous.  Consequently, the resulting grammar
			must be ambiguous.  I'm shutting this warning off.
		 */
		options {
			greedy=false;  // make it exit upon "*/"
			generateAmbigWarnings=false; // shut off newline errors
		}
	:	'\r' '\n'	{newline();}
	|	'\r'		{newline();}
	|	'\n'		{newline();}
	|	~('\n'|'\r')
	)*
	"*/"
	;

OPEN_ELEMENT_OPTION
	:	'<'
	;

CLOSE_ELEMENT_OPTION
	:	'>'
	;

COMMA : ',';

QUESTION :	'?' ;

TREE_BEGIN : "#(" ;

LPAREN:	'(' ;

RPAREN:	')' ;

COLON :	':' ;

STAR:	'*' ;

PLUS:	'+' ;

ASSIGN : '=' ;

IMPLIES : "=>" ;

SEMI:	';' ;

CARET : '^' ;

BANG : '!' ;

OR	:	'|' ;

WILDCARD : '.' ;

RANGE : ".." ;

NOT_OP :	'~' ;

RCURLY:	'}'	;

CHAR_LITERAL
	:	'\'' (ESC|~'\'') '\''
	;

STRING_LITERAL
	:	'"' (ESC|~'"')* '"'
	;

protected
ESC	:	'\\'
		(	'n'
		|	'r'
		|	't'
		|	'b'
		|	'f'
		|	'w'
		|	'a'
		|	'"'
		|	'\''
		|	'\\'
		|	('0'..'3')
			( options { warnWhenFollowAmbig = false; }: '0'..'7'
			( options { warnWhenFollowAmbig = false; }: '0'..'7'
		   )?
			)?
		|	('4'..'7')
			( options { warnWhenFollowAmbig = false; }: '0'..'7' )?
		|	'u' XDIGIT XDIGIT XDIGIT XDIGIT
		)
	;

protected
DIGIT
	:	'0'..'9'
	;

protected
XDIGIT :
		'0' .. '9'
	|	'a' .. 'f'
	|	'A' .. 'F'
	;

INT	:	('0'..'9')+
	;

ARG_ACTION
   :
	NESTED_ARG_ACTION
	{ setText(StringUtils.stripFrontBack(getText(), "[", "]")); }
	;

protected
NESTED_ARG_ACTION :
	'['
	(
		/*	'\r' '\n' can be matched in one alternative or by matching
			'\r' and then '\n' in the next iteration.
		 */
		options {
			generateAmbigWarnings=false; // shut off newline errors
		}
	:	NESTED_ARG_ACTION
	|	'\r' '\n'	{newline();}
	|	'\r'		{newline();}
	|	'\n'		{newline();}
	|	CHAR_LITERAL
	|	STRING_LITERAL
	|	~']'
	)*
	']'
	;

ACTION!
{int actionLine=getLine(); int actionColumn = getColumn(); }
	:	NESTED_ACTION
		(	'?'	{_ttype = SEMPRED;} )?
		{
			if ( _ttype==ACTION ) {
				setText(StringUtils.stripFrontBack(getText(), "{", "}"));
			}
			else {
				setText(StringUtils.stripFrontBack(getText(), "{", "}?"));
			}
			CommonToken t = new CommonToken(_ttype,$getText);
			t.setLine(actionLine);			// set action line to start
			t.setColumn(actionColumn);
			$setToken(t);
		}
	;

protected
NESTED_ACTION :
	'{'
	(
		options {
			greedy = false; // exit upon '}'
		}
	:
		(
			options {
				generateAmbigWarnings = false; // shut off newline warning
			}
		:	'\r' '\n'	{newline();}
		|	'\r' 		{newline();}
		|	'\n'		{newline();}
		)
	|	NESTED_ACTION
	|	CHAR_LITERAL
	|	COMMENT
	|	STRING_LITERAL
	|	.
	)*
	'}'
   ;

TOKEN_REF
options { testLiterals = true; }
	:	'A'..'Z'
		(	// scarf as many letters/numbers as you can
			options {
				warnWhenFollowAmbig=false;
			}
		:
			'a'..'z'|'A'..'Z'|'_'|'0'..'9'
		)*
	;

// we get a warning here when looking for options '{', but it works right
RULE_REF
{
	int t=0;
}
	:	t=INTERNAL_RULE_REF {_ttype=t;}
		(	{t==LITERAL_options}? WS_LOOP ('{' {_ttype = OPTIONS;})?
		|	{t==LITERAL_tokens}? WS_LOOP ('{' {_ttype = TOKENS;})?
		|
		)
	;

protected
WS_LOOP
	:	(	// grab as much WS as you can
			options {
				greedy=true;
			}
		:
			WS
		|	COMMENT
		)*
	;

protected
INTERNAL_RULE_REF returns [int t]
{
	t = RULE_REF;
}
	:	'a'..'z'
		(	// scarf as many letters/numbers as you can
			options {
				warnWhenFollowAmbig=false;
			}
		:
			'a'..'z'|'A'..'Z'|'_'|'0'..'9'
		)*
		{t = testLiteralsTable(t);}
	;

protected
WS_OPT :
	(WS)?
	;
